/* Copyright (c) 2004 Miguel Bazdresch

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

Please see the file LICENSE for more details. */

/* $Id: coder.c,v 1.1 2004/05/02 22:06:19 miguel Rel miguel $ */

/* Miguel Bazdresch */

/* MIMO Simulation */

/* Coder function */
/* This function encodes the bits generated by the source
and frames them according to instructions in mimosim.ini file.
The function returns a 0 if everything went alright, and
returns different than zero if there was an error.
*/

int coder (void) {
    
    /* function prototypes */
    void debug16(void);
    
    /* variables declaration */
    unsigned int j, row, column, tspointer, infopointer;
    
    /* code information bits */
    switch (cChannelCode) {
        case 0:
            /* cChannelCode = 0 : no coding */
            /* Since no coding is needed, just point to original source bits */
            pCodedBits = pSourceBits;
            break;
	default:
	    return 1;
    }
    /* Generate training sequence */
    switch(cTypeTrainingSeq) {
        case 0:
            /* Alternating 1s and 0s */
            pTrainSequence[1] = 1;
            for(j=2;j<=iNumberTrainBitsFrame;j++) {
                pTrainSequence[j] = pTrainSequence[j-1] ? 0 : 1;
            }
            break;
	default:
	    return 1;
    }
    /* Generate coded and framed matrix. This matrix contains the data bits to
       be sent by each antenna in its rows, so that row i contains the vector to
       be transmitted by antenna i. The matrix has M rows and
       (Lt+L)*IBitsPerSymbol columns */
    /* Fill matrix. Fill column by column. */
    column = 1;
    tspointer = 1;
    infopointer = 1;
    /* training sequence columns - all columns are equal */
    for(j=1;j<=(Lt*iBitsPerSymbol);j++) {
        for(row=1;row<=M;row++) {
            pTransmitMatrix[row][column] = pTrainSequence[tspointer];
	    tspointer++;
        }
        column++;
    }
    /* coded sequence columns */
    for(j=1;j<=(L*iBitsPerSymbol);j++) {
        for(row=1;row<=M;row++) {
            pTransmitMatrix[row][column] = pCodedBits[infopointer];
            infopointer++;
        }
        column++;
    }
    if(debug && ((cDebug & 16) == 16))  /* output coded matrix to out file */
        debug16();
    return 0;
}
